# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client / management tool."""

from abc import ABC, abstractmethod
from typing import TypeVar, Type, Tuple, Any, List

import logging
import click

from turbinia_client.helpers import click_helper
from turbinia_client.core.commands import create_request

T = TypeVar('T', bound='FactoryInterface')

log = logging.getLogger('turbinia')


class FactoryInterface(ABC):
  """Factory Interface."""

  @classmethod
  def get_evidence_attributes(
      cls: Type[T], evidence_mapping: dict) -> List[Tuple[str, Any, dict]]:
    """Retrieves a list of evidence attribute metadata.

    Args:
      evidence_mapping (dict): A dictionary of Evidence types
          and attributes retrieved from the Turbinia API server.

    Returns:
      list: A list of 3-tuples comprising the necessary parameters to create
          new click.Option objects.
    """
    map_types = dict(str=str, int=int)
    options_meta = []
    for evidence_name, options in evidence_mapping.items():
      for option_name, option_parameters in options.items():
        param_decls = (['--' + option_name], option_name)
        param_type = map_types.get(option_parameters.get('type'))
        attrs = dict(
            required=option_parameters.get('required'), type=param_type)
        options_meta.append((evidence_name, param_decls, attrs))
    return options_meta

  @classmethod
  @abstractmethod
  def create_dynamic_objects(
      cls: Type[T], name: str, evidence_mapping: dict,
      request_options: dict) -> None:
    """Creates one or more click.Command or click.Options objects.
    
    Args:
      evidence_mapping (dict): A dictionary of Evidence types
          and attributes retrieved from the Turbinia API server.
      request_options (dict): A dictionary of BaseRequestOptions
          attributes and types retrieved from the Turbinia APi server.
    """
    raise NotImplementedError

  @classmethod
  @abstractmethod
  def create_object(
      cls: Type[T], name: str, params: Tuple[Any, dict], callback: Any):
    """Creates a click.Command or click.Options object.

    Args:
      name (str): Object name.
      params (tuple): A tuple comprised of arguments needed to create
          click.Option or click.Command objects.
      callback: A callback function.
    """
    raise NotImplementedError


class OptionFactory(FactoryInterface):
  """Command line options factory class."""

  @classmethod
  def append_request_option_objects(
      cls: Type[T], params: List[click.Option],
      request_options: dict = None) -> None:
    """Appends options to a list of parameters for a click.Command object."""
    for request_option in list(request_options.keys()):
      request_parameters = click_helper.generate_option_parameters(
          request_option)
      params.append(OptionFactory.create_object(params=request_parameters))

  @classmethod
  def create_dynamic_objects(
      cls, name: str = None, evidence_mapping: dict = None,
      request_options: dict = None) -> List[click.Option]:
    """Creates a list of click.Option objects."""
    option_objects = []
    if not name:
      log.info('An evidence type was not provided.')
      return []
    options = OptionFactory.get_evidence_attributes(evidence_mapping)
    for evidence_name, param_decls, attrs in options:
      if name == evidence_name:
        log.debug('Creating option for {0:s}'.format(name))
        click_option = OptionFactory.create_object(params=(param_decls, attrs))
        option_objects.append(click_option)
    return option_objects

  @classmethod
  def create_object(
      cls: Type[T], name: str = None, params: Tuple[Any, dict] = None,
      callback=None) -> click.Option:
    """Creates a click.Option object with specified parameters."""
    option = None
    if params:
      (param_decls, attrs) = params
      option = click.Option(*param_decls, **attrs)
    return option


class CommandFactory(FactoryInterface):
  """Commands factory class."""

  @classmethod
  def create_dynamic_objects(
      cls, name: str = None, evidence_mapping: dict = None,
      request_options: dict = None) -> List[click.Command]:
    """Creates a list of click.Command objects."""

    command_objects = []
    for evidence_name in list(evidence_mapping.keys()):
      log.debug('Creating command for {0:s}'.format(evidence_name))
      # Use lowercase evidence names.
      evidence_name_lower = evidence_name.lower()
      params = OptionFactory.create_dynamic_objects(
          name=evidence_name, evidence_mapping=evidence_mapping,
          request_options=request_options)
      OptionFactory.append_request_option_objects(params, request_options)
      cmd = CommandFactory.create_object(
          name=evidence_name_lower, params=params, callback=create_request)
      command_objects.append(cmd)
    return command_objects

  @classmethod
  def create_object(
      cls: Type[T], name: str = None, params: Tuple[Any, dict] = None,
      callback=None) -> click.Command:
    """Creates a click.Command object with specified parameters."""
    return click.Command(name=name, params=params, callback=callback)



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import click


@click.group('config')
def config_group():
  """Get Turbinia configuration."""


@click.group('status')
def status_group():
  """Get Turbinia request or task status."""


@click.group('result')
def result_group():
  """Get Turbinia request or task results."""


@click.group('jobs')
def jobs_group():
  """Get a list of enabled Turbinia jobs."""


@click.group('submit')
def submit_group():
  """Submit new requests to the Turbinia API server."""



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import logging
import os
import sys

from google_auth_oauthlib import flow
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
from google.auth import exceptions as google_exceptions

log = logging.getLogger('turbinia')


def get_oauth2_credentials(credentials_path, client_secrets_path):
  """Authenticates the user using Google OAuth services."""
  scopes = ['openid', 'https://www.googleapis.com/auth/userinfo.email']
  credentials = None

  # Load credentials file if it exists
  if os.path.exists(credentials_path):
    try:
      credentials = Credentials.from_authorized_user_file(
          credentials_path, scopes)
    except ValueError as exception:
      log.error(f'Error loading credentials: {exception!s}')
    # Refresh credentials using existing refresh_token
    if credentials and credentials.refresh_token:
      log.debug('Found a refresh token. Requesting new id_token...')
      try:
        credentials.refresh(Request())
      except google_exceptions.RefreshError as exception:
        log.error(f'Error refreshing credentials: {exception!s}')
  else:
    # No credentials file, acquire new credentials from secrets file.
    log.info('Could not find existing credentials. Requesting new tokens.')
    try:
      appflow = flow.InstalledAppFlow.from_client_secrets_file(
          client_secrets_path, scopes)
    except FileNotFoundError as exception:
      log.error(f'Client secrets file not found: {exception!s}')
      sys.exit(1)

    log.info(
        'Starting local HTTP server on localhost:8888 for OAUTH flow. '
        'If running turbinia-client remotely over SSH you will need to tunnel '
        'port 8888.')
    appflow.run_local_server(host='localhost', port=8888)
    credentials = appflow.credentials

    # Save credentials
    with open(credentials_path, 'w', encoding='utf-8') as token:
      token.write(credentials.to_json())

  return credentials.id_token



#!/usr/bin/env python
#
# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for formatting text."""

from __future__ import annotations

from abc import ABC, abstractmethod
from click import echo as click_echo

import logging
import json

log = logging.getLogger('turbinia')


def echo_json(json_data: dict) -> None:
  """Pretty print JSON data."""
  if isinstance(json_data, dict):
    click_echo(json.dumps(json_data, indent=2))


class MarkdownReportComponent(ABC):
  """Components for generating Turbinia request/task
      markdown reports.
  """

  def __init__(self):
    """Instantiates a MarkdownReportComponent object."""
    self._components: list(MarkdownReportComponent) = []
    self._parent: MarkdownReportComponent = None
    self._report: str = None

  @property
  def components(self):
    """Returns the components list."""
    return self._components

  @property
  def report(self):
    """Returns the markdown report text."""
    return self._report

  @report.setter
  def report(self, report):
    self._report = report

  @property
  def parent(self) -> MarkdownReportComponent:
    """Returns the parent object."""
    return self._parent

  @parent.setter
  def parent(self, parent: MarkdownReportComponent):
    self._parent = parent

  def bold(self, text):
    """Formats text as bold in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'**{text.strip():s}**'

  def heading1(self, text):
    """Formats text as heading 1 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'# {text.strip():s}'

  def heading2(self, text):
    """Formats text as heading 2 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'## {text.strip():s}'

  def heading3(self, text):
    """Formats text as heading 3 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'### {text.strip():s}'

  def heading4(self, text):
    """Formats text as heading 4 in Markdown format.

    Args:
        text(string): Text to format

    Return:
        string: Formatted text.
    """
    return f'#### {text.strip():s}'

  def heading5(self, text):
    """Formats text as heading 5 in Markdown format.
     Args:
        text(string): Text to format
     Return:
        string: Formatted text.
    """
    return f'##### {text.strip():s}'

  def bullet(self, text, level=1):
    """Formats text as a bullet in Markdown format.

      Args:
        text(string): Text to format
      Return:
        string: Formatted text.
    """
    return f"{'    ' * (level - 1):s}* {text.strip():s}"

  def code(self, text):
    """Formats text as code in Markdown format.

      Args:
          text(string): Text to format

     Return:
          string: Formatted text.
    """
    return f'`{text.strip():s}`'

  def add(self, component: MarkdownReportComponent) -> None:
    """Adds a MarkdownReportComponent object to the components list.

    This method should additionally set the parent object.
    """
    pass

  def add_components(self, components: list[MarkdownReportComponent]) -> None:
    """Adds multiple MarkdownReportComponent objects to the components list."""
    pass

  def remove(self, component: MarkdownReportComponent) -> None:
    """Removes a MarkdownReportComponent object from the components list.

    This method should set the component's object to None.
    """
    pass

  @abstractmethod
  def generate_markdown(self) -> str:
    pass


class TaskMarkdownReport(MarkdownReportComponent):
  """Turbinia Task markdown report."""

  def __init__(self, request_data: dict = None):
    """Initialize TaskMarkdownReport"""
    super().__init__()
    self._request_data: dict = request_data

  def generate_markdown(self) -> str:
    """Generate a markdown report."""
    report: list[str] = []
    task: dict = self._request_data
    if not task:
      return ''

    try:
      report.append(self.heading2(task.get('name')))
      line = f"{self.bold('Evidence:'):s} {task.get('evidence_name')!s}"
      report.append(self.bullet(line))
      line = f"{self.bold('Status:'):s} {task.get('status'):s}"
      report.append(self.bullet(line))
      report.append(self.bullet(f"Task Id: {task.get('id')!s}"))
      report.append(
          self.bullet(f"Executed on worker {task.get('worker_name')!s}"))
      if task.get('report_data'):
        report.append('')
        report.append(self.heading3('Task Reported Data'))
        report.extend(task.get('report_data').splitlines())
      report.append('')
      report.append(self.heading3('Saved Task Files:'))

      saved_paths = task.get('saved_paths')
      if saved_paths:
        for path in saved_paths:
          report.append(self.bullet(self.code(path)))
          report.append('')
    except TypeError as exception:
      log.warning(f'Error formatting the Markdown report: {exception!s}')

    self.report = '\n'.join(report)
    return self.report


class RequestMarkdownReport(MarkdownReportComponent):
  """Turbinia Request Markdown report."""

  def __init__(self, request_data: dict):
    """Initializes a RequestMarkdownReport object."""
    super().__init__()
    self._request_data: dict = request_data

    tasks = [TaskMarkdownReport(task) for task in request_data.get('tasks')]
    self.add_components(tasks)

  def add(self, component: MarkdownReportComponent) -> None:
    if component:
      self.components.append(component)
      component.parent = self

  def remove(self, component: MarkdownReportComponent) -> None:
    self.components.remove(component)
    component.parent = None

  def add_components(self, components: list[MarkdownReportComponent]) -> None:
    if components:
      for component in components:
        self.components.append(component)
        component.parent = self

  def generate_markdown(self) -> str:
    """Generates a Markdown version of Requests results."""
    report: list[str] = []
    request_dict: dict = self._request_data
    if not request_dict:
      return ''

    try:
      report.append(
          self.heading2(f"Request ID: {request_dict.get('request_id')!s}"))
      report.append(
          self.bullet(
              f"Last Update: {request_dict.get('last_task_update_time')!s}"))
      report.append(
          self.bullet(f"Requester: {request_dict.get('requester')!s}"))
      report.append(self.bullet(f"Reason: {request_dict.get('reason')!s}"))
      report.append(self.bullet(f"Status: {request_dict.get('status')!s}"))
      report.append(
          self.bullet(f"Failed tasks: {request_dict.get('failed_tasks'):d}"))
      report.append(
          self.bullet(f"Running tasks: {request_dict.get('running_tasks'):d}"))
      report.append(
          self.bullet(
              f"Successful tasks: {request_dict.get('successful_tasks'):d}"))
      report.append(
          self.bullet(f"Task Count: {request_dict.get('task_count'):d}"))
      report.append(
          self.bullet(f"Queued tasks: {request_dict.get('queued_tasks'):d}"))
      report.append('')
    except TypeError as exception:
      log.warning(f'Error formatting the Markdown report: {exception!s}')

    for task in self.components:
      report.append(task.generate_markdown())

    self.report = '\n'.join(report)
    return self.report


class SummaryMarkdownReport(MarkdownReportComponent):
  """A markdown report summary of all Turbinia Requests."""

  def __init__(self, requests_summary: list[dict]):
    """Initialize SummaryMarkdownReport."""
    super().__init__()
    self._requests_summary = requests_summary

  def generate_markdown(self) -> str:
    """Generate a Markdown version of Requests summary results."""
    if not self._requests_summary:
      return ''
    report: list[str] = []
    requests_status_list = self._requests_summary.get('requests_status')
    for request_dict in requests_status_list:
      request_report = RequestMarkdownReport(request_dict).generate_markdown()
      report.append(request_report)

    self.report = '\n'.join(report)
    return self.report


class WorkersMarkdownReport(MarkdownReportComponent):
  """A markdown report of all tasks for a specific worker."""

  def __init__(self, request_data: dict):
    super().__init__()
    self._request_data: dict = request_data

  def generate_markdown(self) -> str:
    """Generates a Markdown version of tasks per worker."""
    raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import os
import sys
import logging
import click

from urllib3 import exceptions as urllib3_exceptions

from turbinia_client.turbiniacli import TurbiniaCli
from turbinia_client.core import groups
from turbinia_client.factory import factory

_LOGGER_FORMAT = '%(asctime)s %(levelname)s %(name)s - %(message)s'
logging.basicConfig(format=_LOGGER_FORMAT, level=logging.INFO)
log = logging.getLogger('turbinia')


@click.group(context_settings={
    'help_option_names': ['-h', '--help'],
})
@click.option(
    '--config_instance', '-c', help='A Turbinia instance configuration name.',
    show_default=True, show_envvar=True, type=str,
    default=lambda: os.environ.get('TURBINIA_CONFIG_INSTANCE', 'default'))
@click.option(
    '--config_path', '-p', help='Path to the .turbinia_api_config.json file..',
    show_default=True, show_envvar=True, type=str,
    default=lambda: os.environ.get('TURBINIA_API_CONFIG_PATH', '~'))
@click.pass_context
def cli(ctx: click.Context, config_instance: str, config_path: str) -> None:
  """Turbinia API command-line tool (turbinia-client).
  
  \b                         ***    ***                                       
  \b                          *          *                                      
  \b                     ***             ******                                 
  \b                    *                      *                                
  \b                    **      *   *  **     ,*                                
  \b                      *******  * ********                                  
  \b                             *  * *                                         
  \b                             *  * *                                         
  \b                             %%%%%%                                         
  \b                             %%%%%%                                         
  \b                    %%%%%%%%%%%%%%%       %%%%%%                            
  \b              %%%%%%%%%%%%%%%%%%%%%      %%%%%%%                            
  \b %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  ** *******       
  \b %%                                                   %%  ***************   
  \b %%                                (%%%%%%%%%%%%%%%%%%%  *****  **          
  \b   %%%%%        %%%%%%%%%%%%%%%                                             
                                              
  \b   %%%%%%%%%%                     %%          **             ***              
  \b      %%%                         %%  %%             %%%           %%%%,      
  \b      %%%      %%%   %%%   %%%%%  %%%   %%%   %%  %%%   %%%  %%%       (%%    
  \b      %%%      %%%   %%%  %%%     %%     %%/  %%  %%%   %%%  %%%  %%%%%%%%    
  \b      %%%      %%%   %%%  %%%     %%%   %%%   %%  %%%   %%%  %%% %%%   %%%    
  \b      %%%        %%%%%    %%%       %%%%%     %%  %%%    %%  %%%   %%%%%      

  This command-line tool interacts with Turbinia's API server.

  You can specify the API server location in ~/.turbinia_api_config.json
  """
  ctx.obj = TurbiniaCli(
      config_instance=config_instance, config_path=config_path)

  # Set up the tool based on the configuration file parameters.
  ctx.obj.setup()

  # Build all the commands based on responses from the API server.
  request_commands = factory.CommandFactory.create_dynamic_objects(
      evidence_mapping=ctx.obj.evidence_mapping,
      request_options=ctx.obj.request_options)
  for command in request_commands:
    groups.submit_group.add_command(command)


def main():
  """Initialize the cli application."""

  # Register all command groups.
  cli.add_command(groups.submit_group)
  cli.add_command(groups.config_group)
  cli.add_command(groups.jobs_group)
  cli.add_command(groups.result_group)
  cli.add_command(groups.status_group)

  try:
    cli.main()
  except (ConnectionRefusedError, urllib3_exceptions.MaxRetryError,
          urllib3_exceptions.NewConnectionError) as exception:
    log.error(f'Error connecting to the Turbinia API server: {exception!s}')
    sys.exit(-1)


main()



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import os
import logging
import click
import base64
import tarfile

from turbinia_api_lib import exceptions
from turbinia_api_lib import api_client
from turbinia_api_lib.api import turbinia_requests_api
from turbinia_api_lib.api import turbinia_tasks_api
from turbinia_api_lib.api import turbinia_configuration_api
from turbinia_api_lib.api import turbinia_jobs_api
from turbinia_api_lib.api import turbinia_request_results_api

from turbinia_client.core import groups
from turbinia_client.helpers import formatter

log = logging.getLogger('turbinia')


@groups.config_group.command('list')
@click.pass_context
def get_config(ctx: click.Context) -> None:
  """Gets Turbinia server configuration."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(client)
  try:
    api_response = api_instance.read_config()
    formatter.echo_json(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_config: {exception.body}')


@groups.result_group.command('request')
@click.pass_context
@click.argument('request_id')
def get_request_result(ctx: click.Context, request_id: str) -> None:
  """Gets Turbinia request results / output files."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_request_results_api.TurbiniaRequestResultsApi(client)
  try:
    api_response = api_instance.get_request_output(
        request_id, _preload_content=False, _request_timeout=(30, 30))
    filename = f'{request_id}.tgz'
    click.echo(f'Saving output for request {request_id} to: {filename}')
    # Read the response and save into a local file.
    with open(filename, 'wb') as file:
      for chunk in api_response.read_chunked():
        file.write(chunk)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_output: {exception.body}')
  except OSError as exception:
    log.error(f'Unable to save file: {exception}')
  except (ValueError, tarfile.ReadError, tarfile.CompressionError) as exception:
    log.error(f'Error reading saved results file {filename}: {exception}')


@groups.result_group.command('task')
@click.pass_context
@click.argument('task_id')
def get_task_result(ctx: click.Context, task_id: str) -> None:
  """Gets Turbinia task results / output files."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_request_results_api.TurbiniaRequestResultsApi(client)
  try:
    api_response = api_instance.get_task_output(
        task_id, _preload_content=False, _request_timeout=(30, 30))
    filename = f'{task_id}.tgz'
    click.echo(f'Saving output for task {task_id} to: {filename}')

    # Read the response and save into a local file.
    with open(filename, 'wb') as file:
      for chunk in api_response.read_chunked():
        file.write(chunk)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_task_output: {exception.body}')
  except OSError as exception:
    log.error(f'Unable to save file: {exception}')
  except (ValueError, tarfile.ReadError, tarfile.CompressionError) as exception:
    log.error(f'Error reading saved results file {filename}: {exception}')


@groups.jobs_group.command('list')
@click.pass_context
def get_jobs(ctx: click.Context) -> None:
  """Gets Turbinia jobs list."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_jobs_api.TurbiniaJobsApi(client)
  try:
    api_response = api_instance.read_jobs()
    click.echo(api_response)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling read_jobs: {exception.body}')


@groups.status_group.command('request')
@click.pass_context
@click.argument('request_id')
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_request(ctx: click.Context, request_id: str, json_dump: bool) -> None:
  """Gets Turbinia request status."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  if request_id == 'summary':
    click.echo(
        'Oops! "summary" is not a valid request identifier. '
        'Did you mean to run "turbinia-client status summary" instead?')
    return
  try:
    api_response = api_instance.get_request_status(request_id)
    if json_dump:
      formatter.echo_json(api_response)
    else:
      report = formatter.RequestMarkdownReport(api_response).generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_request_status: {exception.body}')


@groups.status_group.command('workers')
@click.pass_context
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_workers(ctx: click.Context, json_dump: bool) -> None:
  click.echo('Not implemented yet.')


@groups.status_group.command('summary')
@click.pass_context
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_requests_summary(ctx: click.Context, json_dump: bool) -> None:
  """Gets a summary of all Turbinia requests."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  try:
    api_response = api_instance.get_requests_summary()
    if json_dump:
      formatter.echo_json(api_response)
    else:
      report = formatter.SummaryMarkdownReport(api_response).generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_requests_summary: {exception.body}')


@groups.status_group.command('task')
@click.pass_context
@click.argument('task_id')
@click.option(
    '--json_dump', '-j', help='Generates JSON output.', is_flag=True,
    required=False)
def get_task(ctx: click.Context, task_id: str, json_dump: bool) -> None:
  """Gets Turbinia task status."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_tasks_api.TurbiniaTasksApi(client)
  try:
    api_response = api_instance.get_task_status(task_id)
    if json_dump:
      formatter.echo_json(api_response)
    else:
      report = formatter.TaskMarkdownReport(api_response).generate_markdown()
      click.echo(report)
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling get_task_status: {exception.body}')


@click.pass_context
def create_request(ctx: click.Context, *args: int, **kwargs: int) -> None:
  """Creates and submits a new Turbinia request."""
  client: api_client.ApiClient = ctx.obj.api_client
  api_instance = turbinia_requests_api.TurbiniaRequestsApi(client)
  evidence_name = ctx.command.name

  # Normalize the evidence class name from lowercase to the original name.
  evidence_name = ctx.obj.normalize_evidence_name(evidence_name)
  # Build request and request_options objects to send to the API server.
  request_options = list(ctx.obj.request_options.keys())
  request = {'evidence': {'type': evidence_name}, 'request_options': {}}

  if 'googlecloud' in evidence_name:
    api_instance_config = turbinia_configuration_api.TurbiniaConfigurationApi(
        client)
    cloud_provider = api_instance_config.read_config()['CLOUD_PROVIDER']
    if cloud_provider != 'GCP':
      log.error(
          f'The evidence type {evidence_name} is Google Cloud only and '
          f'the configured provider for this Turbinia instance is '
          f'{cloud_provider}.')
      return

  for key, value in kwargs.items():
    # If the value is not empty, add it to the request.
    if kwargs.get(key):
      # Check if the key is for evidence or request_options
      if not key in request_options:
        request['evidence'][key] = value
      elif key in ('jobs_allowlist', 'jobs_denylist'):
        jobs_list = value.split(',')
        request['request_options'][key] = jobs_list
      else:
        request['request_options'][key] = value

  if all(key in request['request_options']
         for key in ('recipe_name', 'recipe_data')):
    log.error('You can only provide one of recipe_data or recipe_name')
    return

  recipe_name = request['request_options'].get('recipe_name')
  if recipe_name:
    if not recipe_name.endswith('.yaml'):
      recipe_name = f'{recipe_name}.yaml'
    # Fallback path for the recipe would be TURBINIA_CLI_CONFIG_PATH/recipe_name
    # This is the same path where the client configuration is loaded from.
    recipe_path_fallback = os.path.expanduser(ctx.obj.config_path)
    recipe_path_fallback = os.path.join(recipe_path_fallback, recipe_name)

    if os.path.isfile(recipe_name):
      recipe_path = recipe_name
    elif os.path.isfile(recipe_path_fallback):
      recipe_path = recipe_path_fallback
    else:
      log.error(f'Unable to load recipe {recipe_name}.')
      return

    try:
      with open(recipe_path, 'r', encoding='utf-8') as recipe_file:
        # Read the file and convert to base64 encoded bytes.
        recipe_bytes = recipe_file.read().encode('utf-8')
        recipe_data = base64.b64encode(recipe_bytes)
    except OSError as exception:
      log.error(f'Error opening recipe file {recipe_path}: {exception}')
      return
    except TypeError as exception:
      log.error(f'Error converting recipe data to Base64: {exception}')
      return
    # We found the recipe file, so we will send it to the API server
    # via the recipe_data parameter. To do so, we need to pop recipe_name
    # from the request so that we only have recipe_data.
    request['request_options'].pop('recipe_name')
    # recipe_data should be a UTF-8 encoded string.
    request['request_options']['recipe_data'] = recipe_data.decode('utf-8')

  # Send the request to the API server.
  try:
    log.info(f'Sending request: {request}')
    api_response = api_instance.create_request(request)
    log.info(f'Received response: {api_response}')
  except exceptions.ApiException as exception:
    log.error(
        f'Received status code {exception.status} '
        f'when calling create_request: {exception.body}')
  except (TypeError, exceptions.ApiTypeError) as exception:
    log.error(f'The request object is invalid. {exception}')



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

from typing import Tuple, Sequence


def generate_option_parameters(
    option_name: str) -> Tuple[Tuple[Sequence[str], str], dict]:
  """Builds click.Option or click.Command arguments based on a given parameter 
      name.
  """
  return ((['--' + option_name], option_name), {'required': False, 'type': str})



# -*- coding: utf-8 -*-
# Copyright 2022 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Turbinia API client command-line tool."""

import os
import sys
import logging
import json
import turbinia_api_lib
from typing import Union

from turbinia_api_lib.api import turbinia_configuration_api

from turbinia_client.helpers import auth_helper

log = logging.getLogger('turbinia')


class TurbiniaCli:
  """Turbinia API client tool.

  Attributes:
    _api_client (turbinia_api_lib.ApiClient): An instance of the ApiClient
        class.
    _config (turbinia_api_lib.Configuration): An instance of the
        Configuration class.
    _config_dict (dict): Contains all the cli configuration keys and values
        from the config file.
    _evidence_mapping (dict): Internal dictionary used to map Evidence object
        names to Click commands.
    _request_options (dict): Internal dictinoary used to map Request options
        to Click options.
    api_server_address (str): URL for the API server.
    api_server_port (int): Port for the API server.
    api_authentication_enabled (bool): If set to True, the client will attempt
        to authenticate.
    client_secrets_path (str): Full path to the client secrets file used for
        authentication.
    config_instance (str): A name defined in the configuration file that holds
        all the configuration options for a specific Turbinia deployment.
    config_path (str): Full path to the directory containing the cli tool's
        configuration file.
    credentials_path (str): Full path to the credentials cache file used for
        re-authentication.
  """

  def __init__(self, config_instance=None, config_path=None):
    self._api_client: turbinia_api_lib.ApiClient = None
    self._config: turbinia_api_lib.Configuration = None
    self._config_dict: dict = {}
    self._evidence_mapping: dict = {}
    self._request_options: dict = {}
    self.api_server_address: str = None
    self.api_server_port: int = None
    self.api_authentication_enabled: str = None
    self.client_secrets_path: str = None
    self.config_instance: str = config_instance
    self.config_path: str = config_path
    self.credentials_path: str = None

  def setup(self) -> None:
    """Sets up necessary attributes and preflight requests to the API server.

    The preflight requests get_evidence_arguments and get_request_options
    are used to dynamically build command options.
    """

    # Initialize attributes from config file.
    self.read_api_configuration()

    if not self.config:
      host = f'{self.api_server_address:s}:{self.api_server_port:d}'
      self.config = self.default_config(host)
    if not self.api_client:
      self.api_client = self.default_api_client(self.config)

    if self.api_authentication_enabled:
      log.info(
          f'Authentication is enabled. Using client_secrets file at: '
          f'{self.client_secrets_path:s} and caching credentials at: '
          f'{self.credentials_path:s}')
      self.config.access_token = auth_helper.get_oauth2_credentials(
          self.credentials_path, self.client_secrets_path)

    log.info(
        f'Using configuration instance name -> {self.config_instance:s}'
        f' with host {self.api_server_address:s}:{self.api_server_port:d}')
    try:
      self.evidence_mapping = self.get_evidence_arguments()
      self.request_options = self.get_request_options()
    except turbinia_api_lib.ApiException as exception:
      log.error(
          'Error while attempting to contact the API server during setup: %s',
          exception)
      sys.exit(-1)

  @property
  def api_client(self):
    """Returns an API client object."""
    return self._api_client

  @api_client.setter
  def api_client(self, api_client):
    self._api_client = api_client

  @property
  def config(self):
    """Returns an API client configuration object."""
    return self._config

  @config.setter
  def config(self, config):
    self._config = config

  @property
  def config_dict(self):
    """Returns an API client configuration object."""
    return self._config_dict

  @config_dict.setter
  def config_dict(self, config_dict):
    self._config_dict = config_dict

  @property
  def evidence_mapping(self):
    """Returns a dictionary that contains Evidence object data."""
    return self._evidence_mapping

  @evidence_mapping.setter
  def evidence_mapping(self, evidence_mapping):
    self._evidence_mapping = evidence_mapping

  @property
  def request_options(self):
    """Returns a dictionary that contains RequestOptions object data."""
    return self._request_options

  @request_options.setter
  def request_options(self, request_options):
    self._request_options = request_options

  def default_api_client(
      self,
      config: turbinia_api_lib.Configuration) -> turbinia_api_lib.ApiClient:
    """Default value for API client instance."""
    return turbinia_api_lib.ApiClient(configuration=config)

  def default_config(self, host: str) -> turbinia_api_lib.Configuration:
    """Default value for API client configuration."""
    return turbinia_api_lib.Configuration(host=host)

  def get_evidence_arguments(self, evidence_name=None) -> dict:
    """Gets arguments for Evidence types."""
    api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(
        self.api_client)
    api_response = None
    if evidence_name:
      api_response = api_instance.get_evidence_attributes_by_name(evidence_name)
    else:
      api_response = api_instance.get_evidence_types()

    self.evidence_mapping: dict = api_response
    return api_response

  def get_request_options(self) -> dict:
    """Gets BaseRequestOptions attributes."""
    api_response = None
    api_instance = turbinia_configuration_api.TurbiniaConfigurationApi(
        self.api_client)
    api_response = api_instance.get_request_options()

    return api_response

  def read_api_configuration(self) -> None:
    """Reads the configuration file to obtain the API server URI."""
    if self.config_path == '~':
      client_config_path = os.path.expanduser('~')
      client_config_path = os.path.join(
          client_config_path, '.turbinia_api_config.json')
    else:
      client_config_path = os.path.join(
          self.config_path, '.turbinia_api_config.json')
    try:
      with open(client_config_path, encoding='utf-8') as config:
        config_data = json.loads(config.read())
        config_dict = config_data.get(self.config_instance)
        if not config_dict:
          log.error(f'Error reading configuration key {self.config_instance:s}')
          sys.exit(-1)
        self.api_server_address = config_dict['API_SERVER_ADDRESS']
        self.api_server_port = config_dict['API_SERVER_PORT']
        self.api_authentication_enabled = config_dict[
            'API_AUTHENTICATION_ENABLED']
        credentials_filename = config_dict['CREDENTIALS_FILENAME']
        client_secrets_filename = config_dict['CLIENT_SECRETS_FILENAME']
        home_path = os.path.expanduser('~')
        self.credentials_path = os.path.join(home_path, credentials_filename)
        self.client_secrets_path = os.path.join(
            home_path, client_secrets_filename)
        self.config_dict = config_dict
    except (IOError, FileNotFoundError) as exception:
      log.error(f'Unable to read the configuration file {exception!s}')
      sys.exit(-1)
    except json.JSONDecodeError as exception:
      log.error(f'Error decoding configuration file: {exception!s}')
      sys.exit(-1)
    except KeyError as exception:
      log.error(f'Required configuration key not found: {exception!s}')
      sys.exit(-1)

  def normalize_evidence_name(self, evidence_name_low: str) -> Union[str, None]:
    """Converts a lowercase evidence name into the proper class name."""
    evidence_name = None
    evidence_name_low = evidence_name_low.lower()
    if evidence_name_low:
      for name in self.evidence_mapping.keys():
        if evidence_name_low == name.lower():
          evidence_name = name
          break

    if not evidence_name:
      log.error(f'Unable to map {evidence_name_low} to a valid Evidence name.')
      sys.exit(-1)

    return evidence_name



"""
Turbinia client command-line tool.

Install via 'pip install turbinia-client'
"""
from setuptools import setup, find_packages
from pathlib import Path

NAME = "turbinia-client"
VERSION = "1.0.2"

REQUIRES = [
    "click",
    "turbinia-api-lib",
]

this_directory = Path(__file__).parent
README = (this_directory / "README.md").read_text()

setup(
    name=NAME, 
    version=VERSION,
    description="Turbinia API Client command-line tool.",
    long_description_content_type="text/markdown",
    long_description=README,
    keywords=[
        "Turbinia Client", "Turbinia", "Turbinia API Server"
    ], 
    python_requires=">=3.6", 
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'turbinia-client=turbinia_client.turbiniacli_tool:main'
        ]
    },
    license='Apache License, Version 2.0',
    url='http://turbinia.plumbing/',
    maintainer='Turbinia development team',
    maintainer_email='turbinia-dev@googlegroups.com'
)
